#ifndef HASH_H
#define HASH_H

#include<stdlib.h>
#include<string.h>
#include<ctype.h>
#include<limits.h>
#include<math.h>
#include<errno.h>
#include"prime.h"

#define resize(size,count,ops) (((size) * ((count) << (1))) / (ops))

typedef unsigned long (* key_fn)(void *element);
typedef int (* match_fn)(void *element);
typedef void *(* hsearch_fn)(void *base,int width,int size,unsigned long key,match_fn match);
typedef int (* fail_fn)(void *element);

unsigned long strsign(char *string);
void memswap(void *loc0,void *loc1,int size);

#define hash(key,size) ((key) % (size))
#define rehash(key,prev,size) hash((key)*((prev) + 1),size)
#define kirehash(key,prev,size) hash((key) + (prev),size)
#define ksirehash(key,prev,time,size) hash((key) + (prev) + (time),size)

int hashes;
#define EHRESIZE_FAIL (-1923)

void * hnew(int *about,int width);
#define hdispose(base) free(base)
void *hsearch(void *base,int width,int size,unsigned long key,match_fn match);
void *kihsearch(void *base,int width,int size,unsigned long key,match_fn match);
void *ksihsearch(void *base,int width,int size,unsigned long key,match_fn match);
void **hresize(void **base,int width,int *size,int newsize,hsearch_fn search,key_fn key,match_fn match,fail_fn fail);

typedef int (* ohash_fn)(unsigned long key,int size);
typedef void (* freee_fn)(void *element);

unsigned long keymin = 0,keymax = UINT_MAX;
int idohash(unsigned long key,int size);

#define LOG2 (.30103)
#define keydivisor(size) ((int)(log((double)size) / LOG2))
unsigned keydiv = 2;
int kdohash(unsigned long key,int size);

struct syn {
        void *element;
        struct syn *next;
};
typedef struct syn *syn_t;

struct shptr {
       int index;
       struct syn *prev, *curr;
};
typedef struct shptr *shptr_t;

syn_t * shnew(int *size);
void shdispose(syn_t *base,int size,freee_fn freee);
shptr_t shsearch(syn_t *base,int size,unsigned long key,ohash_fn ohash,match_fn match,shptr_t ptr);
syn_t *shadd(syn_t *base,int size,unsigned long key,void *element,shptr_t ptr);
syn_t *oshadd(syn_t *base,void *element,shptr_t ptr);
syn_t *shrem(syn_t *base,shptr_t ptr,freee_fn freee);
syn_t *oshrem(syn_t *base,shptr_t ptr,freee_fn freee);
#define SHSEEK_FIRST (0)
#define SHSEEK_NEXT (1)
#define SHSEEK_LAST (2)
shptr_t shseek(syn_t *base,int size,int seek,shptr_t ptr);
syn_t **shresize(syn_t **base,int *size,int newsize,key_fn key,ohash_fn ohash);

#endif
